@@ -139,7 +139,7 @@
   (print-unreadable-object (layout stream :type t :identity t)
     (format stream
             "~@[(ID=~d) ~]for ~S~@[, INVALID=~S~]"
-            (layout-id layout)
+            (layout-id layout #-sb-xc-host nil)
             (layout-proper-name layout)
             (layout-invalid layout))))
 
@@ -432,7 +432,8 @@ between the ~A definition and the ~A definition"
         (let ((super (layout-classoid super-layout)))
           (when (and (eq (classoid-state super) :sealed)
                      (not (get-subclassoid classoid super)))
-            (warn "unsealing sealed class ~S in order to subclass it"
+            (#+sb-xc-host error
+             #-sb-xc-host warn "unsealing sealed class ~S in order to subclass it"
                   (classoid-name super))
             (setf (classoid-state super) :read-only))
           (add-subclassoid super classoid (or destruct-layout layout))))))
@@ -1057,7 +1058,7 @@ between the ~A definition and the ~A definition"
          (complex
           :translation complex
           :inherits (number)
-          :codes (,sb-vm:complex-widetag)
+          :codes (,sb-vm:complex-rational-widetag)
           :prototype-form ,(complex 0 1))
          (complex-single-float
           :translation (complex single-float)
@@ -1170,7 +1171,8 @@ between the ~A definition and the ~A definition"
          (vector-nil
           :translation (vector nil)
           :inherits (vector array sequence)
-          :prototype-form (make-array 0 :element-type 'nil :fill-pointer t))
+          :prototype-form (make-array 0 :element-type 'nil :fill-pointer t)
+          :predicate vector-nil-p)
          ;; This name is imperfect. It should be SIMPLE-RANK1-ARRAY-NIL
          ;; to clearly convey that the dimensions are '(*) and not '*.
          (simple-array-nil
@@ -1284,6 +1286,7 @@ between the ~A definition and the ~A definition"
                    :codes (,(sb-vm:saetp-typecode x))
                    :direct-superclasses (vector simple-array)
                    :inherits (vector simple-array array sequence)
+                   :predicate ,(symbolicate (sb-vm:saetp-primitive-type-name x) "-P")
                    :prototype-form
                    (logically-readonlyize
                     (make-array 0 :element-type ',(sb-vm:saetp-specifier x))))))))))
@@ -1299,7 +1302,7 @@ between the ~A definition and the ~A definition"
                            'error
                            (or (getf (cdr x) :predicate)
                                (sb-c::backend-type-predicate translation)))))
-                (assert predicate)
+                (assert predicate nil "~a" name)
                 ;; destructuring-bind will see the first :translation
                 ;; keyword; we don't need to delete the other one.
                 (list* name :predicate predicate :translation translation (cdr x))))
