@@ -420,16 +420,14 @@
        (cdr result)))))
 
 (defun append2 (x y)
-  (declare (optimize (sb-c:verify-arg-count 0)))
   (if (null x)
       y
-      (let ((result (list (car x))))
-        (do ((more (cdr x) (cdr more))
-             (tail result (cdr tail)))
+      (let* ((result (cons (car x) y))
+             (tail result))
+        (do ((more (cdr x) (cdr more)))
             ((null more)
-             (rplacd (truly-the cons tail) y)
              result)
-                    (rplacd (truly-the cons tail) (list (car more)))))))
+          (rplacd (truly-the cons tail) (setf tail (cons (car more) y)))))))
 
 
 ;;;; list copying functions
@@ -438,6 +436,16 @@
   "Return a new list which is EQUAL to LIST. LIST may be improper."
   (copy-list-macro list))
 
+(defun copy-list-to (list tail)
+  (declare (explicit-check))
+  (the list (cdr (truly-the cons tail)))
+  (do ((orig list (cdr orig))
+       (splice tail
+               (let ((cell (list (car orig))))
+                 (rplacd (truly-the cons splice) cell)
+                 cell)))
+      ((atom orig) (rplacd (truly-the cons splice) orig))))
+
 (defun ensure-heap-list (list)
   (declare (sb-c::tlab :system))
   ;; If some cons is not on the heap then copy the whole list
