@@ -13,6 +13,23 @@
 
 ;;;; structure frobbing primitives
 
+#+permgen
+(defun allocate-permgen-layout (nwords)
+  (flet ((thunk ()
+           (let ((freeptr sb-vm:*permgen-space-free-pointer*))
+             (setf sb-vm:*permgen-space-free-pointer*
+                   ;; round-to-odd, add the header word
+                   (sap+ freeptr (ash (1+ (logior nwords 1)) sb-vm:word-shift)))
+             (aver (<= (sap-int sb-vm:*permgen-space-free-pointer*)
+                       (+ sb-vm:permgen-space-start sb-vm:permgen-space-size)))
+             (setf (sap-ref-word freeptr 0)
+                   (logior (ash nwords sb-vm:instance-length-shift)
+                           sb-vm:instance-widetag))
+             (%make-lisp-obj (sap-int (sap+ freeptr sb-vm:instance-pointer-lowtag))))))
+    (if (sb-thread::mutex-p sb-vm::*allocator-mutex*)
+        (with-system-mutex (sb-vm::*allocator-mutex* :without-gcing t) (thunk))
+        (thunk))))
+
 ;;; For lack of any better to place to write up some detail surrounding
 ;;; layout creation for structure types, I'm putting here.
 ;;; When you issue a DEFSTRUCT at the REPL, there are *three* instances
@@ -44,10 +61,13 @@
          (layout
           (truly-the layout
                      #+compact-instance-header
-                     (sb-vm::alloc-immobile-fixedobj
-                      (1+ nwords)
-                      (logior (ash nwords sb-vm:instance-length-shift)
-                              sb-vm:instance-widetag))
+                     (progn
+                       #+permgen (allocate-permgen-layout nwords)
+                       #+immobile-space
+                       (sb-vm::alloc-immobile-fixedobj
+                        (1+ nwords)
+                        (logior (ash nwords sb-vm:instance-length-shift)
+                                sb-vm:instance-widetag)))
                      #-compact-instance-header
                      (%make-instance/mixed nwords))))
     (%set-instance-layout layout #.(find-layout 'layout))
@@ -166,7 +186,7 @@
                         layout (+ (get-dsd-index layout id-word0) index))
              ;; use SAP-ref, for lack of half-sized slots
              #+64-bit `(signed-sap-ref-32 (id-bits-sap) (ash index 2))))
-(defun layout-id (layout)
+(defun layout-id (layout &optional (assign t))
   ;; If a structure type at depthoid >= 2, then fetch the INDEXth id
   ;; where INDEX is depthoid - 2. Otherwise fetch the 0th id.
   ;; There are a few non-structure types at positive depthoid; those do not store
@@ -180,8 +200,9 @@
          (id (with-pinned-objects (layout)
                (access-it))))
     (truly-the
-     layout-id
-     (cond ((zerop id)
+     (or null layout-id)
+     (cond ((not (zerop id)) id)
+           (assign
             (aver (logior +structure-layout-flag+ (layout-flags layout)))
             (with-system-mutex (*layout-id-mutex*)
               (let ((id (truly-the layout-id (access-it)))) ; double-check
@@ -190,8 +211,7 @@
                       (setf (access-it)
                             ;; doesn't really need ATOMIC- any moren
                             (atomic-incf (car *layout-id-generator*))))
-                    id))))
-           (t id)))))
+                    id))))))))
 
 (defun set-layout-inherits (layout inherits structurep this-id)
   (setf (layout-inherits layout) inherits)
@@ -537,10 +557,8 @@
 ;;; If "flexible" defstructs (multiple inheritance, standard-object ancestors) are ever
 ;;; brought to life, these might be inadmissible. Probably we would not store the
 ;;; fast map in such situations.
-#+64-bit
-(progn
 (defun best-slot-map-parameters (names)
-  (let* ((raw-hashes (mapcar (lambda (x) (sxhash (the symbol x))) names))
+  (let* ((raw-hashes (mapcar (lambda (x) (symbol-hash (the symbol x))) names))
          (n-names (length names))
          (n-cells (logior n-names 1)) ; round to odd if not already
          ;; remember the best N-COLLISIONS N-CELLS MASK C
@@ -582,34 +600,29 @@
             (aref map 2) c)
       (fill map nil :start (+ fast-slot-table-fixed-cells n-cells))
       (dolist (dsd (dd-slots dd) map)
-        ;; minor kludge: skip raw slots for now, and let structure-slot-value
-        ;; fall back to the fully general SLOT-VALUE. A better solution would be
-        ;; to encode the raw-type into the value stored in the index vector.
-        (when (eq (dsd-raw-type dsd) t)
           (binding*
-              ((hash (sxhash (dsd-name dsd)))
-                 (masked-hash (logand (ash hash (- shift)) mask))
-               (bin (truly-the index
-                               (+ (sb-vm::fastrem-32 masked-hash c n-cells)
-                                  fast-slot-table-fixed-cells)))
-               (dsd-index (dsd-index dsd))
-               (name (dsd-name dsd))
-               ((key value)
-                (cond ((eql (svref map bin) 0) ; empty, just store the name and dsd-index
-                       (values name dsd-index))
-                      ;; A bin with a collision is upgraded to a vector of the two entries
-                      ((symbolp (svref map bin))
-                       (values (vector (svref map bin) name)
-                               (vector (svref map (+ bin n-cells)) dsd-index)))
-                      ;; Multiple collisions
-                      (t
-                       (values (concatenate 'vector (svref map bin) (vector name))
-                               (concatenate 'vector
-                                            (svref map (+ bin n-cells))
-                                            (vector dsd-index)))))))
-              (setf (svref map bin) key
-                    (svref map (+ bin n-cells)) value)))))))
-) ; end PROGN
+           ((hash (symbol-hash (dsd-name dsd)))
+            (masked-hash (logand (ash hash (- shift)) mask))
+            (bin (truly-the index
+                            (+ (sb-vm::fastrem-32 masked-hash c n-cells)
+                               fast-slot-table-fixed-cells)))
+            (dsd-bits (dsd-bits dsd))
+            (name (dsd-name dsd))
+            ((key value)
+             (cond ((eql (svref map bin) 0) ; empty, just store the name and dsd-index
+                    (values name dsd-bits))
+                   ;; A bin with a collision is upgraded to a vector of the two entries
+                   ((symbolp (svref map bin))
+                    (values (vector (svref map bin) name)
+                            (vector (svref map (+ bin n-cells)) dsd-bits)))
+                   ;; Multiple collisions
+                   (t
+                    (values (concatenate 'vector (svref map bin) (list name))
+                            (concatenate 'vector
+                                         (svref map (+ bin n-cells))
+                                         (list dsd-bits)))))))
+           (setf (svref map bin) key
+                 (svref map (+ bin n-cells)) value))))))
 
 
 (/show0 "target-defstruct.lisp end of file")
