@@ -307,6 +307,8 @@
           (pushnew :sb-futex sb-xc:*features*))
         (when (target-featurep '(:and :sb-thread :x86-64))
           (pushnew :system-tlabs sb-xc:*features*))
+        (when (target-featurep '(:and :mark-region-gc :permgen :x86-64))
+          (pushnew :compact-instance-header sb-xc:*features*))
         (when (target-featurep :immobile-space)
           (when (target-featurep :x86-64)
             (pushnew :compact-instance-header sb-xc:*features*))
@@ -386,7 +388,7 @@
           ;; It sorta kinda works to have both, but there should be no need,
           ;; and it's not really supported.
           "At most one interpreter can be selected")
-         ("(and compact-instance-header (not immobile-space))"
+         ("(and compact-instance-header (not (or permgen immobile-space)))"
           ":COMPACT-INSTANCE-HEADER requires :IMMOBILE-SPACE feature")
          ("(and immobile-code (not immobile-space))"
           ":IMMOBILE-CODE requires :IMMOBILE-SPACE feature")
@@ -876,6 +878,25 @@
 (compile 'install-read-interceptor)
 
 (defvar *math-ops-memoization* (make-hash-table :test 'equal))
+(defun math-journal-pathname (direction)
+  ;; Initialy we read from the file in the source tree, but writeback occurs
+  ;; to a new local file. Then if re-reading we read the local copy of the cache.
+  ;; This should allow multiple builds to happen (via make-all-targets.sh) in a
+  ;; single source tree. If exactly one target is built, we can mv the local file
+  ;; on top of the source file. For more than one, we could either merge them
+  ;; or just ignore any modifications.
+  (let* ((base "xfloat-math.lisp-expr")
+         (local (concatenate 'string sb-cold::*host-obj-prefix* base)))
+    (pathname
+     (ecase direction
+       (:input (if (probe-file local) local base))
+       (:output local)))))
+
+(defun count-lines-of (pathname &aux (n 0))
+  (with-open-file (f pathname)
+    (loop (let ((line (read-line f nil)))
+            (if line (incf n) (return n))))))
+
 (defmacro with-math-journal (&body body)
   `(let* ((table *math-ops-memoization*)
           (memo (cons table (hash-table-count table))))
@@ -888,14 +909,111 @@
      (when nil ; *compile-verbose*
        (funcall (intern "SHOW-INTERNED-NUMBERS" "SB-IMPL") *standard-output*))
      (when (> (hash-table-count table) (cdr memo))
-       (let ((filename "float-math.lisp-expr"))
+       (let ((filename (math-journal-pathname :output)))
          (with-open-file (stream filename :direction :output
                                           :if-exists :supersede)
            (funcall (intern "DUMP-MATH-MEMOIZATION-TABLE" "SB-IMPL")
                     table stream))
-         (format t "~&; wrote ~a - ~d entries"
+         ;; Enforce absence of spurious newlines from pretty-printing or whatever
+         ;; If this assertion is wrong on other lisps we can just remove it
+         (assert (= (count-lines-of filename) (+ (hash-table-count table) 5)))
+         (format t "~&; Math journal: wrote ~S (~d entries)"
                  filename (hash-table-count table))))))
 
+;;;; One more journal file because the math file isn't enough
+;;; Define this before renaming the SB- packages
+;;; A non-parallelized compile using a sufficiently new SBCL as the host
+;;; will use a paravirtualized implementation of generate-perfect-hash-sexpr
+;;; which is to say, it just uses the host; as a side-effect it records
+;;; the generated string so that we can replay it for any host
+;;; or for parallelized build.
+(defvar *perfect-hash-generator-mode* :PLAYBACK)
+(defvar *perfect-hash-generator-memo* nil)
+(defvar *perfect-hash-generator-journal* "xperfecthash.lisp-expr")
+#+sbcl (when (and (find-symbol "MAKE-PERFECT-HASH-LAMBDA" "SB-C")
+                  (find-symbol "NEWCHARSTAR-STRING" "SB-INT"))
+         (pushnew :use-host-hash-generator cl:*features*)
+         (setq *perfect-hash-generator-mode* :RECORD))
+
+;;; I want this to work using the host-native readtable if sb-cold:*xc-readtable*
+;;; isn't established. The caller should bind *READTABLE* to ours if reading
+;;; on a non-SBCL host; it's purposely not done here.
+(defun preload-perfect-hash-generator (pathname)
+  (with-open-file (stream pathname :if-does-not-exist nil)
+    (when stream
+      (let ((entries (let ((*read-base* 16)) (read stream))))
+        (setq *perfect-hash-generator-memo*
+              ;; Compute the XOR all the hashes of each entry as a quick pass/fail
+              ;; when searching, assuming thst EQUALP compares (CAR CONS) before
+              ;; the CDR, which is almost surely, though not necessarily, what it does.
+              (mapcar (lambda (entry)
+                        (destructuring-bind (array . string) entry
+                          ;; assert that the entry was stored in canonical form
+                          (assert (equalp array (sort (copy-seq array) #'<)))
+                          (let ((digest (reduce #'logxor array)))
+                            (cons (cons digest array) string))))
+                      entries))))))
+
+(defun emulate-generate-perfect-hash-sexpr (array)
+  (let ((computed
+         #+use-host-hash-generator
+         (let ((string
+                (sb-int:newcharstar-string
+                 (sb-sys:with-pinned-objects (array)
+                   (sb-alien:alien-funcall
+                    (sb-alien:extern-alien
+                     "generate_perfhash_sexpr"
+                     (function (* sb-alien:char) sb-alien:system-area-pointer sb-alien:int))
+                    (sb-sys:vector-sap array) (length array))))))
+           ;; don't need the final newline, it looks un-lispy in the file
+           (let ((l (length string)))
+             (assert (char= (char string (1- l)) #\newline))
+             (subseq string 0 (1- l))))))
+    ;; Entries are written to disk with hashes sorted in ascending order so that
+    ;; comparing as sets can be done using EQUALP.
+    ;; Sort nondestructively in case something else looks at the value as supplied.
+    (let* ((canonical-array (sort (copy-seq array) #'<))
+           (digest (reduce #'logxor canonical-array))
+           (match (assoc (cons digest canonical-array) *perfect-hash-generator-memo*
+                         :test #'equalp)))
+      (when match
+        (when computed (assert (string= (cdr match) computed)))
+        (return-from emulate-generate-perfect-hash-sexpr (cdr match)))
+    (ecase *perfect-hash-generator-mode*
+      (:playback
+       (error "perfect hash file is missing a needed entry for ~x" array))
+      (:record
+       (setf *perfect-hash-generator-memo*
+             (nconc *perfect-hash-generator-memo*
+                    (list (cons (cons digest canonical-array) computed))))
+       computed)))))
+
+(defun compile-perfect-hashfun-for-host (lambda)
+  ;; Remove the declare:
+  ;;   (declare (optimize (safety 0) (sb-c:store-source-form 0)))
+  (let ((third (third lambda)))
+    (assert (eq (car third) 'declare))
+    (let ((new `(lambda ,(second lambda) ,@(cdddr lambda))))
+      (compile nil new))))
+(defun maybe-save-perfect-hashfuns-for-playback ()
+  ;; FIXME: file corruption occurs for -jN > 1 from make-all-targets.sh
+  #+use-host-hash-generator
+  (when (eq *perfect-hash-generator-mode* :record)
+    (with-open-file (stream *perfect-hash-generator-journal*
+                            :direction :output
+                            :if-existS :supersede :if-does-not-exist :create)
+      (write-char #\( stream)
+      (dolist (entry *perfect-hash-generator-memo*)
+        (destructuring-bind ((digest . array) . string) entry
+          (declare (ignore digest))
+          (write (cons array string):stream stream :length nil :base 16
+                 :pretty t :right-margin 128))
+        (terpri stream))
+      (write-string ")
+;; EOF
+" stream)))
+  t)
+
 ;;;; Please avoid writing "consecutive" (un-nested) reader conditionals
 ;;;; in this file, whether for the same or different feature test.
 ;;;; The following example prints 3 different results in 3 different lisp
