@@ -81,7 +81,7 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
 ;;; Verify that compile-time floating-point math matches load-time.
 (defvar *compile-files-p*)
 (when (or (not (boundp '*compile-files-p*)) *compile-files-p*)
-  (with-open-file (stream "float-math.lisp-expr" :if-does-not-exist nil)
+  (with-open-file (stream "xfloat-math.lisp-expr" :if-does-not-exist nil)
     (when stream
       (format t "; Checking ~S~%" (pathname stream))
       ;; Ensure that we're reading the correct variant of the file
@@ -90,14 +90,17 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
       (sb-kernel::with-float-traps-masked (:overflow :divide-by-zero)
         (let ((*readtable* (copy-readtable))
               (*package* (find-package "SB-KERNEL")))
+          ;; The reasoning behind this limited-use variant of read-time-eval is that
+          ;; since it is too early to actually use EVAL in the interpreted fashion,
+          ;; EVAL would call COMPILE which is just ridiculous because it would mean
+          ;; compiling however many #. expression there are in this file
           (set-dispatch-macro-character
            #\# #\. (lambda (stream subchar arg)
                      (declare (ignore subchar arg))
                      (let ((expr (read stream t nil t)))
                        (ecase (car expr)
-                         (sb-kernel:make-single-float
-                          (sb-kernel:make-single-float (second expr)))
-                         (sb-kernel:make-double-float
+                         (sb-kernel::s (sb-kernel:make-single-float (second expr)))
+                         (sb-kernel::d
                           (sb-kernel:make-double-float (second expr) (third expr)))))))
           (dolist (expr (read stream))
             (destructuring-bind (fun args . result) expr
@@ -150,6 +153,12 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                  (read f) ; skip over the make-host-{1,2} input files
                  (read f)))
       (sb-c::*handled-conditions* sb-c::*handled-conditions*))
+ ;; The CONCATENATE transform involves REPLACE which involves UBn-BASH-COPY which involves
+ ;; SHIFT-TOWARDS-{START|END} which is called with a constant arg. But the interpreter stubs
+ ;; aren't compiled yet. So in attempting to constant-fold the call, CAREFUL-CALL gets an
+ ;; undefined-fun error, which is handled fine unless you've broken the handler for undefined-fun
+ ;; and are trying to debug it in early warm load.
+ (declare (notinline concatenate))
  (proclaim '(sb-ext:muffle-conditions compiler-note))
  (flet ((do-srcs (list)
          (dolist (stem list)
@@ -217,6 +226,7 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                          ;; rather than asking what to do here?
                          #+(or x86 x86-64) ;; these should complete without warnings
                          (cerror "Ignore warnings" "Compile completed with warnings")))
+                  #+nil (sb-impl::show-hash-cache-statistics)
                   (unless (handler-bind
                               ((sb-kernel:redefinition-with-defgeneric
                                 #'muffle-warning))
@@ -247,7 +257,7 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
 
 (sb-c::dump/restore-interesting-types 'write)
 (when (hash-table-p sb-c::*static-vop-usage-counts*)
-  (with-open-file (output "output/warm-vop-usage.txt"
+  (with-open-file (output (merge-pathnames "warm-vop-usage.txt" *objfile-prefix*)
                           :direction :output :if-exists :supersede)
     (let (list)
       (sb-int:dohash ((name vop) sb-c::*backend-parsed-vops*)
