@@ -52,6 +52,7 @@
 ;;; zero-length, non-zero-total-size vector-with-fill-pointer.
 (declaim (type (and (vector t) (not simple-array)) *constraint-universe*))
 (defvar *constraint-universe*)
+(defvar *blocks-to-terminate*)
 
 (defstruct (vector-length-constraint
             (:constructor make-vector-length-constraint (var))
@@ -92,9 +93,10 @@
 (defstruct (equality-constraint
             (:include constraint)
             (:constructor make-equality-constraint
-                (number operator x y not-p
+                (number operator x y not-p &optional (amount 0)
                  &aux (kind 'equality))))
-  (operator nil :type symbol))
+  (operator nil :type symbol)
+  (amount 0 :type integer))
 
 ;;; The basic interval type. It can handle open and closed intervals.
 ;;; A bound is open if it is a list containing a number, just like
@@ -400,6 +402,11 @@
         (when (lambda-var-p y)
           (register-constraint y new x))
         new)))
+
+(declaim (inline type-for-constraints-p))
+(defun type-for-constraints-p (type)
+  (not (or (eq type *universal-type*)
+           (contains-hairy-type-p type))))
 
 ;;; Actual conset interface
 ;;;
@@ -589,7 +596,14 @@
                   (conset-lvar-lambda-var-eql-p constraints lvar lambda-var)
                   lambda-var)))
           ((cast-p use)
-           (ok-lvar-lambda-var (cast-value use) constraints)))))
+           (ok-lvar-lambda-var (cast-value use) constraints))
+          ((set-p use)
+           (let ((leaf (set-var use)))
+             (when (and (lambda-var-p leaf)
+                        (lambda-var-constraints leaf))
+               (and leaf
+                    (conset-lvar-lambda-var-eql-p constraints lvar leaf)
+                    leaf)))))))
 ;;;; Searching constraints
 
 ;;; Add the indicated test constraint to TARGET.
@@ -1000,7 +1014,8 @@
                     (setq not-type (type-union not-type other)))))
             (eql
              (let ((other-type (leaf-type other)))
-               (cond ((not not-p)
+               (cond ((not (type-for-constraints-p other-type)))
+                     ((not not-p)
                       (intersect-result other-type))
                      ((constant-p other)
                       (cond ((member-type-p other-type)
@@ -1117,36 +1132,38 @@
                (kind (constraint-kind con)))
           (case kind
             (typep
-             (if not-p
-                 (if (member-type-p other)
-                     (mapc-member-type-members #'note-not other)
-                     (setq not-res (type-union not-res other)))
-                 (setq res (type-intersection res other))))
+             (when (type-for-constraints-p other)
+               (if not-p
+                   (if (member-type-p other)
+                       (mapc-member-type-members #'note-not other)
+                       (setq not-res (type-union not-res other)))
+                   (setq res (type-intersection res other)))))
             (eql
              (let ((other-type (leaf-type other)))
-               (if not-p
-                   (when (constant-p other)
-                     (cond ((member-type-p other-type)
-                            (note-not (constant-value other)))
-                           ;; Numeric types will produce interesting
-                           ;; negations, other than just "not equal"
-                           ;; which can be handled by the equality
-                           ;; constraints.
-                           ((numeric-type-p other-type)
-                            (add-to-xset (constant-value other) not-numeric))))
-                   (let ((leaf-type (leaf-type leaf)))
-                     (cond
-                       ((or (constant-p other)
-                            (and (leaf-refs other) ; protect from
+               (when (type-for-constraints-p other-type)
+                 (if not-p
+                     (when (constant-p other)
+                       (cond ((member-type-p other-type)
+                              (note-not (constant-value other)))
+                             ;; Numeric types will produce interesting
+                             ;; negations, other than just "not equal"
+                             ;; which can be handled by the equality
+                             ;; constraints.
+                             ((numeric-type-p other-type)
+                              (add-to-xset (constant-value other) not-numeric))))
+                     (let ((leaf-type (leaf-type leaf)))
+                       (cond
+                         ((or (constant-p other)
+                              (and (leaf-refs other) ; protect from
                                         ; deleted vars
-                                 (csubtypep other-type leaf-type)
-                                 (not (type= other-type leaf-type))
-                                 ;; Don't change to a LEAF not visible here.
-                                 (leaf-visible-from-node-p other ref)))
-                        (change-ref-leaf ref other)
-                        (when (constant-p other) (return)))
-                       (t
-                        (setq res (type-intersection res other-type))))))))
+                                   (csubtypep other-type leaf-type)
+                                   (not (type= other-type leaf-type))
+                                   ;; Don't change to a LEAF not visible here.
+                                   (leaf-visible-from-node-p other ref)))
+                          (change-ref-leaf ref other)
+                          (when (constant-p other) (return)))
+                         (t
+                          (setq res (type-intersection res other-type)))))))))
             ((< > <= >= =)
              (let ((type (type-after-comparison kind not-p res y)))
                (when type
@@ -1181,7 +1198,13 @@
                    (t
                     (derive-node-type ref
                                       (make-single-value-type type))
-                    (maybe-terminate-block ref nil)))))))
+                    (when (eq (node-derived-type ref) *empty-type*)
+                      ;; Terminating blocks early may leave loops with
+                      ;; just one entry point, resulting in
+                      ;; monotonically growing variables without a
+                      ;; starting point which will propagate new
+                      ;; constraints for each increment.
+                      (pushnew ref *blocks-to-terminate*))))))))
   (values))
 
 ;;;; Flow analysis
@@ -1220,7 +1243,7 @@
                  and val in (combination-args call)
                  when (and val (lambda-var-constraints var))
                  do (let ((type (lvar-type val)))
-                      (unless (eq type *universal-type*)
+                      (when (type-for-constraints-p type)
                         (conset-add-constraint gen 'typep var type nil)))
                     (maybe-add-eql-var-var-constraint var val gen)
                     (add-var-result-constraints var val gen)))))
@@ -1233,9 +1256,9 @@
        (let* ((lvar (cast-value node))
               (var (ok-lvar-lambda-var lvar gen))
               (atype (single-value-type (cast-derived-type node))))
-         (cond ((eq atype *universal-type*))
-               (var
-                (conset-add-constraint-to-eql gen 'typep var atype nil)))
+         (when (and var
+                    (type-for-constraints-p atype))
+           (conset-add-constraint-to-eql gen 'typep var atype nil))
          (constraint-propagate-back lvar 'typep atype gen gen nil)
          (when (and (bound-cast-p node)
                     (bound-cast-check node)
@@ -1252,17 +1275,24 @@
          (when (policy node (or (and (= speed 3) (> speed compilation-speed))
                                 (> debug 1)))
            (let ((type (lambda-var-type var)))
-             (unless (eq *universal-type* type)
+             (when (type-for-constraints-p type)
                (do-eql-vars (other (var gen))
                  (unless (eql other var)
                    (conset-add-constraint gen 'typep other type nil))))))
-         (conset-clear-lambda-var gen var)
+
+         (let ((new (add-set-constraints var (set-value node) gen)))
+           (conset-clear-lambda-var gen var)
+           (when new
+             (conset-union gen new)))
+
          (let ((type (single-value-type (node-derived-type node))))
-           (unless (eq type *universal-type*)
+           (when (type-for-constraints-p type)
              (conset-add-constraint gen 'typep var type nil)))
          (unless (policy node (> compilation-speed speed))
            (maybe-add-eql-var-var-constraint var (set-value node) gen))
-         (add-eq-constraint var (set-value node) gen)))
+         (add-eq-constraint var (set-value node) gen)
+         (when (node-lvar node)
+           (conset-add-lvar-lambda-var-eql gen (node-lvar node) var))))
       (combination
        (when (eq (combination-kind node) :known)
          (unless (and preprocess-refs-p
@@ -1385,8 +1415,9 @@
 
 ;;; Join the constraints coming from the predecessors of BLOCK on
 ;;; every constrained variable into the constraint set IN.
-(defun join-type-constraints (in block)
+(defun join-type-constraints (in block &optional equality-only)
   (let ((vars '())
+        (equality-vars)
         (predecessors (block-pred block)))
     (dolist (pred predecessors)
       (let ((out (block-out-for-successor pred block)))
@@ -1395,12 +1426,18 @@
             (let ((kind  (constraint-kind con))
                   (y     (constraint-y con))
                   (not-p (constraint-not-p con)))
-              (when (or (member kind '(typep < >))
-                        (and (eq kind 'eql) (or (not not-p)
-                                                (constant-p y)))
-                        (and (eq kind '=) (and (numeric-type-p y)
-                                               (not not-p))))
-                (pushnew (constraint-x con) vars))))
+              (when (and (not equality-only)
+                         (or (member kind '(typep < >))
+                             (and (eq kind 'eql) (or (not not-p)
+                                                     (constant-p y)))
+                             (and (eq kind '=) (and (numeric-type-p y)
+                                                    (not not-p)))))
+                (pushnew (constraint-x con) vars))
+              (when (and (eq kind 'equality)
+                         (/= (equality-constraint-amount con) 0))
+                (pushnew (constraint-x con) equality-vars)
+                (when (lambda-var/vector-length-p y)
+                  (pushnew y equality-vars)))))
           (return))))
     (dolist (var vars)
       (let ((in-var-type *empty-type*))
@@ -1411,7 +1448,7 @@
                     (type-union in-var-type
                                 (type-from-constraints var out *universal-type*))))))
 
-        (unless (eq in-var-type *universal-type*)
+        (when (type-for-constraints-p in-var-type)
           ;; Remove the existing constraints to avoid joining them again later.
           (do-propagatable-constraints (con (in var))
             (when (eq (constraint-kind con) 'typep)
@@ -1420,7 +1457,9 @@
                                                     var
                                                     in-var-type
                                                     nil)
-                         in))))))
+                         in))))
+    (dolist (var equality-vars)
+      (join-equality-constraints var block in))))
 
 (defun compute-block-in (block join-types-p)
   (let ((in nil))
@@ -1432,8 +1471,8 @@
           (if in
               (conset-intersection in out)
               (setq in (copy-conset out))))))
-    (when (and join-types-p (rest (block-pred block)))
-      (join-type-constraints in block))
+    (when (rest (block-pred block))
+      (join-type-constraints in block (not join-types-p)))
     (or in (make-conset))))
 
 (defun update-block-in (block join-types-p)
@@ -1512,9 +1551,10 @@
           ;; Propagate constraints
           (loop for block = (pop blocks-to-process)
                 while block do
-                  (unless (eq block (component-tail component))
-                    (when (update-block-in block join-types-p)
-                      (enqueue (find-block-type-constraints block nil))))))
+                (unless (or (block-delete-p block)
+                            (eq block (component-tail component)))
+                  (when (update-block-in block join-types-p)
+                    (enqueue (find-block-type-constraints block nil))))))
 
         rest-of-blocks))))
 
@@ -1529,8 +1569,10 @@
         (setf (if-alternative-constraints last) nil)
         (setf (if-consequent-constraints last) nil))))
   (setf (block-out (component-head component)) (make-conset))
-  (dolist (block (find-and-propagate-constraints component))
-    (unless (block-delete-p block)
-      (use-result-constraints block)))
-
+  (let (*blocks-to-terminate*)
+    (dolist (block (find-and-propagate-constraints component))
+      (unless (block-delete-p block)
+        (use-result-constraints block)))
+    (loop for node in *blocks-to-terminate*
+          do (maybe-terminate-block node nil)))
   (values))
