@@ -26,18 +26,33 @@
        n
        (%single-float n)))
 
-(deftransform %single-float ((n) (single-float) *)
+(deftransform %single-float ((n) (single-float) * :important nil)
   'n)
 
-(deftransform %double-float ((n) (double-float) *)
+(deftransform %double-float ((n) (double-float) * :important nil)
   'n)
 
-(deftransform %single-float ((n) (ratio) *)
+(deftransform %single-float ((n) (ratio) * :important nil)
   '(sb-kernel::single-float-ratio n))
 
-(deftransform %double-float ((n) (ratio) *)
+(deftransform %double-float ((n) (ratio) * :important nil)
   '(sb-kernel::double-float-ratio n))
 
+(macrolet ((def (type from-type)
+             `(deftransform ,(symbolicate "%" type) ((n) ((or ,type ,from-type)) * :important nil)
+                (when (or (csubtypep (lvar-type n) (specifier-type ',type))
+                          (csubtypep (lvar-type n) (specifier-type ',from-type)))
+                  (give-up-ir1-transform))
+                `(if (,',(symbolicate type "-P") n)
+                     (truly-the ,',type n)
+                     (,',(symbolicate "%" type) (truly-the ,',from-type n))))))
+  (def single-float double-float)
+  (def single-float sb-vm:signed-word)
+  (def single-float word)
+  (def double-float single-float)
+  (def double-float sb-vm:signed-word)
+  (def double-float word))
+
 ;;; RANDOM
 (macrolet ((frob (fun type)
              `(deftransform random ((num &optional state)
@@ -465,9 +480,11 @@
 
 (macrolet ((def (name prim rtype)
              `(progn
-               (deftransform ,name ((x) (single-float) ,rtype)
-                 `(coerce (,',prim (coerce x 'double-float)) 'single-float))
-               (deftransform ,name ((x) (double-float) ,rtype)
+               (deftransform ,name ((x) (single-float) ,rtype :node node)
+                 (delay-ir1-transform node :ir1-phases)
+                 `(%single-float (,',prim (%double-float x))))
+               (deftransform ,name ((x) (double-float) ,rtype :node node)
+                 (delay-ir1-transform node :ir1-phases)
                  `(,',prim x)))))
   (def exp %exp *)
   (def log %log float)
@@ -759,7 +776,7 @@
 (defun safe-expt (x y)
   (when (and (numberp x) (numberp y))
     (handler-case
-        (when (< (abs y) 10000)
+        (when (sb-xc:< (abs y) 10000)
           (expt x y))
       ;; Currently we can hide unanticipated errors (such as failure to use SB-XC: math
       ;; when cross-compiling) as well as the anticipated potential problem of overflow.
@@ -1126,11 +1143,18 @@
 (defoptimizer (phase derive-type) ((num))
   (one-arg-derive-type num #'phase-derive-type-aux #'phase))
 
-(deftransform realpart ((x) ((complex rational)) *)
+(deftransform realpart ((x) ((complex rational)) * :important nil)
   '(%realpart x))
-(deftransform imagpart ((x) ((complex rational)) *)
+(deftransform imagpart ((x) ((complex rational)) * :important nil)
   '(%imagpart x))
 
+(deftransform realpart ((x) (real) * :important nil)
+  'x)
+(deftransform imagpart ((x) ((and single-float (not (eql $-0f0)))) * :important nil)
+  $0f0)
+(deftransform imagpart ((x) ((and double-float (not (eql $-0d0)))) * :important nil)
+  $0d0)
+
 ;;; Make REALPART and IMAGPART return the appropriate types. This
 ;;; should help a lot in optimized code.
 (defun realpart-derive-type-aux (type)
@@ -1260,47 +1284,47 @@
                 ;; they be implemented on an all-or-none basis.
                 (unless (vop-existsp :named sb-vm::%negate/complex-double-float)
                 ;; negation
-                (deftransform %negate ((z) ((complex ,type)) *)
+                (deftransform %negate ((z) ((complex ,type)) * :important nil)
                   '(complex (%negate (realpart z)) (%negate (imagpart z))))
                 ;; complex addition and subtraction
-                (deftransform + ((w z) ((complex ,type) (complex ,type)) *)
+                (deftransform + ((w z) ((complex ,type) (complex ,type)) * :important nil)
                   '(complex (+ (realpart w) (realpart z))
                             (+ (imagpart w) (imagpart z))))
-                (deftransform - ((w z) ((complex ,type) (complex ,type)) *)
+                (deftransform - ((w z) ((complex ,type) (complex ,type)) * :important nil)
                   '(complex (- (realpart w) (realpart z))
                             (- (imagpart w) (imagpart z))))
                 ;; Add and subtract a complex and a real.
-                (deftransform + ((w z) ((complex ,type) real) *)
+                (deftransform + ((w z) ((complex ,type) real) * :important nil)
                   `(complex (+ (realpart w) z)
                             (+ (imagpart w) ,(coerce 0 ',type))))
-                (deftransform + ((z w) (real (complex ,type)) *)
+                (deftransform + ((z w) (real (complex ,type)) * :important nil)
                   `(complex (+ (realpart w) z)
                             (+ (imagpart w) ,(coerce 0 ',type))))
                 ;; Add and subtract a real and a complex number.
-                (deftransform - ((w z) ((complex ,type) real) *)
+                (deftransform - ((w z) ((complex ,type) real) * :important nil)
                   `(complex (- (realpart w) z)
                             (- (imagpart w) ,(coerce 0 ',type))))
-                (deftransform - ((z w) (real (complex ,type)) *)
+                (deftransform - ((z w) (real (complex ,type)) * :important nil)
                   `(complex (- z (realpart w))
                             (- ,(coerce 0 ',type) (imagpart w))))
                 ;; Multiply a complex by a real or vice versa.
-                (deftransform * ((w z) ((complex ,type) real) *)
+                (deftransform * ((w z) ((complex ,type) real) * :important nil)
                   '(complex (* (realpart w) z) (* (imagpart w) z)))
-                (deftransform * ((z w) (real (complex ,type)) *)
+                (deftransform * ((z w) (real (complex ,type)) * :important nil)
                   '(complex (* (realpart w) z) (* (imagpart w) z)))
                 ;; conjugate of complex number
-                (deftransform conjugate ((z) ((complex ,type)) *)
+                (deftransform conjugate ((z) ((complex ,type)) * :important nil)
                   '(complex (realpart z) (- (imagpart z))))
                 ;; comparison
-                (deftransform = ((w z) ((complex ,type) (complex ,type)) *)
+                (deftransform = ((w z) ((complex ,type) (complex ,type)) * :important nil)
                   '(and (= (realpart w) (realpart z))
                     (= (imagpart w) (imagpart z))))
-                (deftransform = ((w z) ((complex ,type) real) *)
+                (deftransform = ((w z) ((complex ,type) real) * :important nil)
                   '(and (= (realpart w) z) (zerop (imagpart w))))
-                (deftransform = ((w z) (real (complex ,type)) *)
+                (deftransform = ((w z) (real (complex ,type)) * :important nil)
                   '(and (= (realpart z) w) (zerop (imagpart z))))
                 ;; Multiply two complex numbers.
-                (deftransform * ((x y) ((complex ,type) (complex ,type)) *)
+                (deftransform * ((x y) ((complex ,type) (complex ,type)) * :important nil)
                   '(let* ((rx (realpart x))
                           (ix (imagpart x))
                           (ry (realpart y))
@@ -1308,12 +1332,12 @@
                     (complex (- (* rx ry) (* ix iy))
                              (+ (* rx iy) (* ix ry)))))
                 ;; Divide a complex by a real.
-                (deftransform / ((w z) ((complex ,type) real) *)
+                (deftransform / ((w z) ((complex ,type) real) * :important nil)
                   '(complex (/ (realpart w) z) (/ (imagpart w) z)))
                 )
 
                 ;; Divide two complex numbers.
-                (deftransform / ((x y) ((complex ,type) (complex ,type)) *)
+                (deftransform / ((x y) ((complex ,type) (complex ,type)) * :important nil)
                   (if (vop-existsp :translate sb-vm::swap-complex)
                       '(let* ((cs (conjugate (sb-vm::swap-complex x)))
                               (ry (realpart y))
@@ -1339,7 +1363,7 @@
                               (complex (/ (+ (* rx r) ix) dn)
                                        (/ (- (* ix r) rx) dn)))))))
                 ;; Divide a real by a complex.
-                (deftransform / ((x y) (real (complex ,type)) *)
+                (deftransform / ((x y) (real (complex ,type)) * :important nil)
                   (if (vop-existsp :translate sb-vm::swap-complex)
                       '(let* ((ry (realpart y))
                               (iy (imagpart y)))
@@ -1392,53 +1416,53 @@
 
 (flet ((def (op)
          (%deftransform op nil '(function (single-float real) single-float)
-                        #'single-float-real-contagion)
+                        #'single-float-real-contagion nil)
          (%deftransform op nil '(function (real single-float) single-float)
-                        #'real-single-float-contagion)
+                        #'real-single-float-contagion nil)
          (%deftransform op nil '(function (double-float real))
-                        #'double-float-real-contagion)
+                        #'double-float-real-contagion nil)
          (%deftransform op nil '(function (real double-float))
-                        #'real-double-float-contagion)
+                        #'real-double-float-contagion nil)
 
          (%deftransform op nil '(function ((complex single-float) real) (complex single-float))
-                        #'single-float-real-contagion)
+                        #'single-float-real-contagion nil)
          (%deftransform op nil '(function (real (complex single-float)) (complex single-float))
-                        #'real-single-float-contagion)
+                        #'real-single-float-contagion nil)
          (%deftransform op nil '(function ((complex double-float) real) (complex double-float))
-                        #'double-float-real-contagion)
+                        #'double-float-real-contagion nil)
          (%deftransform op nil '(function (real (complex double-float)) (complex double-float))
-                        #'real-double-float-contagion)))
+                        #'real-double-float-contagion nil)))
   (dolist (op '(+ * / -))
     (def op)))
 
 (flet ((def (op)
          (%deftransform op nil `(function (single-float (integer ,most-negative-exactly-single-float-integer
                                                                  ,most-positive-exactly-single-float-integer)))
-                        #'single-float-real-contagion)
+                        #'single-float-real-contagion nil)
          (%deftransform op nil `(function ((integer ,most-negative-exactly-single-float-integer
                                                     ,most-positive-exactly-single-float-integer)
                                            single-float))
-                        #'real-single-float-contagion)
+                        #'real-single-float-contagion nil)
 
          (%deftransform op nil `(function (double-float
                                            (or single-float
                                                (integer ,most-negative-exactly-double-float-integer
                                                         ,most-positive-exactly-double-float-integer))))
-                        #'double-float-real-contagion)
+                        #'double-float-real-contagion nil)
          (%deftransform op nil `(function ((or single-float
                                                (integer ,most-negative-exactly-double-float-integer
                                                         ,most-positive-exactly-double-float-integer))
                                            double-float))
-                        #'real-double-float-contagion)))
+                        #'real-double-float-contagion nil)))
   (dolist (op '(= < > <= >=))
     (def op)))
 
 (%deftransform '= nil '(function ((complex double-float) single-float))
-               #'double-float-real-contagion)
+               #'double-float-real-contagion nil)
 (%deftransform '= nil '(function (single-float (complex double-float)))
-               #'real-double-float-contagion)
+               #'real-double-float-contagion nil)
 
-(deftransform complex ((realpart &optional imagpart) (rational &optional (or null (integer 0 0))))
+(deftransform complex ((realpart &optional imagpart) (rational &optional (or null (integer 0 0))) * :important nil)
   'realpart)
 
 ;;; Here are simple optimizers for SIN, COS, and TAN. They do not
@@ -1640,17 +1664,33 @@
              ,(value-within-numeric-type rem-type))))
 
 (macrolet ((def (type)
-             `(deftransform unary-truncate ((number) (,type))
-                '(if (typep number
-                      '(,type
-                        ,(symbol-value (package-symbolicate :sb-kernel 'most-negative-fixnum- type))
-                        ,(symbol-value (package-symbolicate :sb-kernel 'most-positive-fixnum- type))))
-                  (let ((truncated (truly-the fixnum (,(symbolicate '%unary-truncate/ type) number))))
-                    (declare (flushable ,(symbolicate "%" type)))
-                    (values truncated
-                            (- number
-                               (coerce truncated ',type))))
-                  (,(symbolicate 'unary-truncate- type '-to-bignum) number)))))
+             `(deftransform unary-truncate ((number) (,type) * :node node)
+                (let ((cast (cast-or-check-bound-type node)))
+                  (if (and cast
+                           (csubtypep cast (specifier-type 'sb-vm:signed-word)))
+                      (let ((int (type-approximate-interval cast)))
+                        (when int
+                          (multiple-value-bind (low high) (,(package-symbolicate :sb-kernel type '-integer-bounds)
+                                                           (interval-low int)
+                                                           (interval-high int))
+                            `(if (typep number
+                                        '(,',type ,low ,high))
+                                 (let ((truncated (truly-the ,(type-specifier cast) (,',(symbolicate '%unary-truncate/ type) number))))
+                                   (declare (flushable ,',(symbolicate "%" type)))
+                                   (values truncated
+                                           (- number
+                                              (coerce truncated ',',type))))
+                                 ,(internal-type-error-call 'number (type-specifier cast) 'truncate-to-integer)))))
+                      '(if (typep number
+                            '(,type
+                              ,(symbol-value (package-symbolicate :sb-kernel 'most-negative-fixnum- type))
+                              ,(symbol-value (package-symbolicate :sb-kernel 'most-positive-fixnum- type))))
+                        (let ((truncated (truly-the fixnum (,(symbolicate '%unary-truncate/ type) number))))
+                          (declare (flushable ,(symbolicate "%" type)))
+                          (values truncated
+                                  (- number
+                                     (coerce truncated ',type))))
+                        (,(symbolicate 'unary-truncate- type '-to-bignum) number)))))))
   (def single-float)
   (def double-float))
 
@@ -1850,7 +1890,7 @@
     (declare (optimize speed (safety 0)))
     (let* ((high (double-float-high-bits x))
            (low (double-float-low-bits x))
-           (exp (ldb sb-vm:double-float-exponent-byte high))
+           (exp (ldb sb-vm:double-float-hi-exponent-byte high))
            (biased (the double-float-exponent
                         (- exp sb-vm:double-float-bias))))
       (declare (type (signed-byte 32) high)
