@@ -157,6 +157,7 @@
 ;;; represented by a call to VALUES.
 (defknown values (&rest t) * (movable flushable))
 (defknown values-list (list) * (movable foldable unsafely-flushable))
+(defknown reverse-values-list (list index) * (movable foldable unsafely-flushable))
 
 ;;;; from the "Macros" chapter:
 
@@ -282,6 +283,10 @@
     (fixnum real fixnum) boolean
     (foldable flushable movable no-verify-arg-count))
 
+(defknown (check-range< check-range<= check-range<<= check-range<=<)
+    (fixnum t fixnum) boolean
+    (foldable flushable movable no-verify-arg-count))
+
 (defknown (two-arg-gcd two-arg-lcm two-arg-and two-arg-ior two-arg-xor two-arg-eqv)
   (integer integer) integer
   (no-verify-arg-count))
@@ -356,11 +361,10 @@
 (defknown unary-truncate (real) (values integer real)
   (movable foldable flushable no-verify-arg-count))
 
-;;; KLUDGE: Don't fold these, the compiler may call it on floats that
-;;; do not truncate into a bignum, and the functions do not check
-;;; their input values and produce corrupted memory.
 (defknown unary-truncate-single-float-to-bignum (single-float) (values bignum (eql $0f0))
-    (#+(or) foldable movable flushable fixed-args))
+    (foldable movable flushable fixed-args)
+  :folder #'truncate)
+
 (defknown unary-truncate-double-float-to-bignum (double-float)
     (values #+64-bit bignum #-64-bit integer
             (and
@@ -368,12 +372,15 @@
                     (not (or riscv ppc64))) ;; they can't survive cold-init
              (eql $0d0)
              double-float))
-   (#+(or) foldable movable flushable fixed-args))
+   (foldable movable flushable fixed-args)
+  :folder #'truncate)
 
 (defknown %unary-truncate-single-float-to-bignum (single-float) bignum
-   (#+(or) foldable movable flushable fixed-args))
+   (foldable movable flushable fixed-args)
+  :folder (lambda (float) (values (truncate float))))
 (defknown %unary-truncate-double-float-to-bignum (double-float) bignum
-   (#+(or) foldable movable flushable fixed-args))
+   (foldable movable flushable fixed-args)
+  :folder (lambda (float) (values (truncate float))))
 
 (defknown (subtract-bignum add-bignums) (bignum bignum) integer
     (movable flushable no-verify-arg-count))
@@ -512,19 +519,16 @@
 ;; generated by the compiler, and successive args checked by hand.
 ;; The case-insensitive functions don't need any checks, since the underlying
 ;; two-arg case-insensitive function does it, except when it isn't called.
-(defknown (char=)
+(defknown (char= char-equal)
   (character &rest character) boolean (movable foldable flushable commutative))
 
 (defknown (char/= char< char> char<= char>= char-not-equal)
   (character &rest character) boolean (movable foldable flushable))
-(defknown (char-equal char-lessp char-greaterp char-not-greaterp char-not-lessp)
+(defknown (char-lessp char-greaterp char-not-greaterp char-not-lessp)
   (character &rest character) boolean (movable foldable flushable))
 
-(defknown (two-arg-char-equal)
-    (character character) boolean
-  (movable foldable flushable commutative no-verify-arg-count))
-
-(defknown (two-arg-char-not-equal
+(defknown (two-arg-char-equal
+           two-arg-char-not-equal
            two-arg-char-lessp
            two-arg-char-not-lessp
            two-arg-char-greaterp
@@ -575,6 +579,9 @@
   :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t
                                           :preserve-vector-type t))
 
+(defknown list-reverse-into-vector (proper-list) simple-vector
+  (flushable no-verify-arg-count))
+
 (defknown make-sequence (type-specifier index
                                         &key
                                         (:initial-element t))
@@ -586,17 +593,29 @@
   :derive-type (creation-result-type-specifier-nth-arg 0))
 
 (defknown %concatenate-to-string (&rest sequence) simple-string
-  (flushable no-verify-arg-count))
+  (flushable))
 (defknown %concatenate-to-base-string (&rest sequence) simple-base-string
-  (flushable no-verify-arg-count))
+  (flushable))
 (defknown %concatenate-to-list (&rest sequence) list
-    (flushable no-verify-arg-count))
+    (flushable))
 (defknown %concatenate-to-simple-vector (&rest sequence) simple-vector
-  (flushable no-verify-arg-count))
+  (flushable))
 (defknown %concatenate-to-vector ((unsigned-byte #.sb-vm:n-widetag-bits) &rest sequence)
     vector
   (flushable no-verify-arg-count))
 
+(defknown %concatenate-to-string-subseq (&rest t) simple-string
+  (flushable))
+(defknown %concatenate-to-base-string-subseq (&rest t) simple-base-string
+  (flushable))
+(defknown %concatenate-to-list-subseq (&rest t) list
+    (flushable))
+(defknown %concatenate-to-simple-vector-subseq (&rest t) simple-vector
+  (flushable))
+(defknown %concatenate-to-vector-subseq ((unsigned-byte #.sb-vm:n-widetag-bits) &rest t)
+    vector
+  (flushable no-verify-arg-count))
+
 (defknown (possibly-base-stringize possibly-base-stringize-to-heap) (string) simple-string
   (flushable no-verify-arg-count))
 
@@ -995,6 +1014,8 @@
   :call-type-deriver #'append-call-type-deriver)
 
 (defknown copy-list (proper-or-dotted-list) list (flushable))
+(defknown sb-impl::copy-list-to (proper-or-dotted-list cons) cons
+  (flushable no-verify-arg-count))
 (defknown copy-alist (proper-list) list (flushable))
 (defknown copy-tree (t) t (flushable recursive))
 (defknown revappend (proper-list t) t (flushable))
@@ -1919,7 +1940,7 @@
 (defknown %dynamic-extent-start () t)
 (defknown %special-bind (t t) t)
 (defknown %special-unbind (&rest symbol) t)
-(defknown %listify-rest-args (t index) list (flushable))
+(defknown %listify-rest-args (t index) list (flushable always-translatable))
 (defknown %more-arg-context (t t) (values t index) (flushable))
 (defknown %more-arg (t index) t (flushable))
 (defknown %more-kw-arg (t fixnum) (values t t))
