@@ -90,8 +90,9 @@
                            ;; table, which works nicely assuming a register is wired to the card table
                            #+(and sb-safepoint (not x86-64))
                            ;; Must be just before NIL.
-                           (safepoint ,(symbol-value '+backend-page-bytes+) gc-safepoint-page-addr)
+                           (safepoint ,(symbol-value '+backend-page-bytes+))
                            (static ,small-space-size)
+                           (permgen 8388608) ; 8MiB
                            #+darwin-jit
                            (static-code ,small-space-size))
                          #+immobile-space
@@ -103,7 +104,9 @@
            (loop for (space size var-name) in spaces
                  appending
                  (let* ((relocatable
-                          (member space '(fixedobj text #+relocatable-static-space static
+                          (member space '(fixedobj text permgen
+                                          #+relocatable-static-space safepoint
+                                          #+relocatable-static-space static
                                           #+immobile-space alien-linkage-table
                                           read-only)))
                         (start ptr)
@@ -371,7 +374,7 @@
   (defconstant +pseudo-static-generation+ 6))
 
 (defparameter *runtime-asm-routines* nil)
-(defparameter *linkage-space-predefined-entries* nil)
+(defparameter *alien-linkage-table-predefined-entries* nil)
 
 ;;; Floating-point related constants, both format descriptions and FPU
 ;;; control register descriptions.  These don't exactly match up with
