@@ -176,18 +176,16 @@
      (%instance-set .instance. .index. .newval.)
      .newval.))
 
-;;; *** These transforms should be the only code, aside from the C runtime
-;;;     with knowledge of the layout index.
 #+compact-instance-header
 (define-source-transform function-with-layout-p (x) `(functionp ,x))
 #-compact-instance-header
 (progn
+  (define-source-transform function-with-layout-p (x) `(funcallable-instance-p ,x))
+  ;; Nothing but these transforms should assume that slot 0 holds a layout
   (define-source-transform %instance-layout (x)
     `(truly-the layout (%instance-ref ,x 0)))
   (define-source-transform %set-instance-layout (instance layout)
-    `(%instance-set ,instance 0 (the layout ,layout)))
-  (define-source-transform function-with-layout-p (x)
-    `(funcallable-instance-p ,x)))
+    `(%instance-set ,instance 0 (the layout ,layout))))
 
 ;;;; simplifying HAIRY-DATA-VECTOR-REF and HAIRY-DATA-VECTOR-SET
 
@@ -308,20 +306,24 @@
          (element-ctype (array-type-upgraded-element-type type))
          (declared-element-ctype (declared-array-element-type type)))
     (declare (type ctype element-ctype))
-    (when (eq *wild-type* element-ctype)
-      (give-up-ir1-transform
-       "Upgraded element type of array is not known at compile time."))
-    (let ((element-type-specifier (type-specifier element-ctype)))
-      `(multiple-value-bind (array index)
-           (%data-vector-and-index array index)
-         (declare (type (simple-array ,element-type-specifier 1) array)
-                  (type ,element-type-specifier new-value))
-         ,(if (type= element-ctype declared-element-ctype)
-              '(progn (data-vector-set array index new-value)
-                      new-value)
-              `(progn (data-vector-set array index
-                       ,(the-unwild declared-element-ctype 'new-value))
-                      ,(truly-the-unwild declared-element-ctype 'new-value)))))))
+    (cond ((eq *wild-type* element-ctype)
+           ;; The new value is only suitable for a simple-vector
+           (if (csubtypep (lvar-type new-value) (specifier-type '(not (or number character))))
+               `(hairy-data-vector-set (the simple-vector array) index new-value)
+               (give-up-ir1-transform
+                "Upgraded element type of array is not known at compile time.")))
+          (t
+           (let ((element-type-specifier (type-specifier element-ctype)))
+             `(multiple-value-bind (array index)
+                  (%data-vector-and-index array index)
+                (declare (type (simple-array ,element-type-specifier 1) array)
+                         (type ,element-type-specifier new-value))
+                ,(if (type= element-ctype declared-element-ctype)
+                     '(progn (data-vector-set array index new-value)
+                       new-value)
+                     `(progn (data-vector-set array index
+                                              ,(the-unwild declared-element-ctype 'new-value))
+                             ,(truly-the-unwild declared-element-ctype 'new-value)))))))))
 
 ;;; Transform multi-dimensional array to one dimensional data vector
 ;;; access.
@@ -591,14 +593,16 @@
 
 ;;;; transforms for EQL of floating point values
 (unless (vop-existsp :named sb-vm::eql/single-float)
-(deftransform eql ((x y) (single-float single-float))
-  '(= (single-float-bits x) (single-float-bits y))))
+(deftransform eql ((x y) (single-float single-float) * :node node)
+  (delay-ir1-transform node :ir1-phases)
+  '(eql (single-float-bits x) (single-float-bits y))))
 
 (unless (vop-existsp :named sb-vm::eql/double-float)
-(deftransform eql ((x y) (double-float double-float))
-  #-64-bit '(and (= (double-float-low-bits x) (double-float-low-bits y))
-                  (= (double-float-high-bits x) (double-float-high-bits y)))
-  #+64-bit '(= (double-float-bits x) (double-float-bits y))))
+(deftransform eql ((x y) (double-float double-float) * :node node)
+  (delay-ir1-transform node :ir1-phases)
+  #-64-bit '(and (eql (double-float-low-bits x) (double-float-low-bits y))
+             (eql (double-float-high-bits x) (double-float-high-bits y)))
+  #+64-bit '(eql (double-float-bits x) (double-float-bits y))))
 
 
 ;;;; modular functions
