@@ -1106,13 +1106,16 @@
                                              (vop-block vop)
                                              (template-or-lose 'move)
                                              (reference-tn widetag nil)
-                                             (reference-tn (tn-ref-tn (vop-results vop )) t)
+                                             (reference-tn (tn-ref-tn (vop-results vop)) t)
                                              vop)
                         (let ((next (vop-next vop)))
                           (when (and next
                                      (eq (vop-name next) 'branch-if))
                             (setf info (vop-codegen-info next))
                             (delete-vop next))
+                          (when (and (eql (car tags) sb-vm:simple-array-widetag)
+                                     (csubtypep (tn-ref-type (vop-args vop)) (specifier-type 'string)))
+                            (pop tags))
                           (emit-and-insert-vop (vop-node vop)
                                                (vop-block vop)
                                                test-vop
@@ -1289,6 +1292,13 @@
     (or (ir2-block-%label block)
         (setf (ir2-block-%label block) (gen-label)))))
 
+(defun next-vop-label (vop)
+  (let* ((block (vop-block vop))
+         (next (ir2-block-next block)))
+    (when (eq (ir2-block-last-vop block) vop)
+      (or (ir2-block-%label next)
+          (setf (ir2-block-%label next) (gen-label))))))
+
 (when-vop-existsp (:named sb-vm::signed-byte-64-p-move-to-word)
   (flet ((opt (vop new-vop &optional not-vop)
            (let ((dest (branch-destination vop))
@@ -1416,7 +1426,7 @@
 (defglobal *vops-allowed-within-pseudo-atomic*
     '(set-slot %raw-instance-set/word %raw-instance-set/signed-word %raw-instance-set/single
       %raw-instance-set/double %raw-instance-set/complex-single %raw-instance-set/complex-double
-      move move-operand make-unbound-marker make-funcallable-instance-tramp
+      move move-operand make-unbound-marker
       sb-vm::move-from-word/fixnum sb-vm::move-to-word/fixnum
       sb-vm::move-from-fixnum+1 sb-vm::move-from-fixnum-1))
 
