@@ -38,9 +38,6 @@
 ;;; the alloc granularity, it will, once we are smarter about finding
 ;;; the start of objects.
 (defconstant gencgc-alloc-granularity 0)
-;;; The minimum size at which we release address ranges to the OS.
-;;; This must be a multiple of the OS page size.
-(defconstant gencgc-release-granularity +backend-page-bytes+)
 ;;; The card size for immobile/low space
 (defconstant immobile-card-bytes 4096)
 
@@ -91,7 +88,7 @@
 #+(or linux darwin)
 (gc-space-setup #x50000000
                      :read-only-space-size 0
-                     :fixedobj-space-size #.(* 40 1024 1024)
+                     :fixedobj-space-size #.(* 60 1024 1024)
                      :text-space-size #.(* 130 1024 1024)
                      :dynamic-space-start #x1000000000)
 
@@ -163,7 +160,7 @@
   #'equalp)
 
 (defconstant-eqx +static-fdefns+
-    `#(ensure-symbol-hash sb-impl::install-hash-table-lock update-object-layout
+    `#(sb-impl::install-hash-table-lock update-object-layout
        ,@common-static-fdefns)
   #'equalp)
 
