@@ -210,12 +210,15 @@
           (let ((header (car remaining))
                 (last (null (cdr remaining))))
             (cond
-             ((atom header)
-              (inst cmp :byte widetag-tn header)
-              (if last
-                  (inst jmp equal target)
-                  (inst jmp :e when-true)))
-             (t
+              ((and (eql header simple-array-widetag)
+                    value-tn-ref
+                    (csubtypep (tn-ref-type value-tn-ref) (specifier-type 'string))))
+              ((atom header)
+               (inst cmp :byte widetag-tn header)
+               (if last
+                   (inst jmp equal target)
+                   (inst jmp :e when-true)))
+              (t
                (let ((start (car header))
                      (end (cdr header)))
                  (cond
@@ -329,42 +332,59 @@
     (inst test value value)))
 
 (macrolet ((define (name src-size)
-             `(progn
-                (define-vop (,name)
-                  (:translate ,name)
-                  (:args (value :scs (any-reg descriptor-reg)))
-                  (:conditional :z)
-                  (:arg-refs arg-ref)
-                  (:policy :fast-safe)
-                  (:temporary (:sc unsigned-reg) temp temp2)
-                  (:generator 6
-                    ;; Optimistically assume that the argument is a fixnum.
-                    ;; If it is, then we're shifting out the tag bit, otherwise
-                    ;; putting random bits in the low byte, which is harmless.
-                    (move temp value)
-                    (inst sar temp n-fixnum-tag-bits)
-                    (inst movsx '(,src-size :qword) temp2 temp)
-                    (inst cmp temp2 temp)
-                    ;; If the input wasn't known to be a fixnum, and it passed
-                    ;; the sign-extension test, then check FIXNUMP now.
-                    (unless (csubtypep (tn-ref-type arg-ref) (specifier-type 'fixnum))
-                      (inst jmp :nz OUT) ; did NOT pass the sign-extension test
-                      (inst test :byte value fixnum-tag-mask))
-                    OUT))
-                (define-vop (,(symbolicate name "/SIGNED"))
-                  (:translate ,name)
-                  (:args (value :scs (signed-reg)))
-                  (:arg-types signed-num)
-                  (:conditional :z)
-                  (:policy :fast-safe)
-                  (:temporary (:sc unsigned-reg) temp)
-                  (:generator 2
-                    (inst movsx '(,src-size :qword) temp value)
-                    (inst cmp temp value))))))
+             `(define-vop (,(symbolicate name "/SIGNED"))
+                (:translate ,name)
+                (:args (value :scs (signed-reg)))
+                (:arg-types signed-num)
+                (:conditional :z)
+                (:policy :fast-safe)
+                (:temporary (:sc unsigned-reg) temp)
+                (:generator 2
+                  (inst movsx '(,src-size :qword) temp value)
+                  (inst cmp temp value)))))
   (define signed-byte-8-p :byte)
   (define signed-byte-16-p :word)
   (define signed-byte-32-p :dword))
 
+(define-vop (signed-byte-8-p)
+  (:translate signed-byte-8-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 6
+    (inst lea temp (ea (ash (expt 2 7) n-fixnum-tag-bits) value))
+    (inst test temp (lognot (fixnumize (1- (expt 2 8)))))))
+
+(define-vop (signed-byte-16-p)
+  (:translate signed-byte-16-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 6
+    (inst lea temp (ea (ash (expt 2 15) n-fixnum-tag-bits) value))
+    (inst test temp (lognot (fixnumize (1- (expt 2 16)))))))
+
+(define-vop (signed-byte-32-p)
+  (:translate signed-byte-32-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp temp2)
+  (:generator 6
+    (move temp value)
+    (inst sar temp n-fixnum-tag-bits)
+    (inst movsx '(:dword :qword) temp2 temp)
+    (inst cmp temp2 temp)
+    (unless (csubtypep (tn-ref-type arg-ref) (specifier-type 'fixnum))
+      (inst jmp :nz out)
+      (inst test :byte value fixnum-tag-mask))
+    out))
+
 ;;; Sign bit and fixnum tag bit.
 (defconstant non-negative-fixnum-mask-constant
   #x8000000000000001)
@@ -443,66 +463,6 @@
 
         (emit-label not-target)))))
 
-(defun power-of-two-limit-p (x)
-  (and (fixnump x)
-       (= (logcount (1+ x)) 1)))
-
-(define-vop (test-fixnum-mod-power-of-two)
-  (:args (value :scs (any-reg descriptor-reg
-                              unsigned-reg signed-reg)))
-  (:arg-types *
-              (:constant (satisfies power-of-two-limit-p)))
-  (:translate fixnum-mod-p)
-  (:conditional :e)
-  (:info hi)
-  (:policy :fast-safe)
-  (:arg-refs value-ref)
-  (:vop-var vop)
-  (:generator 4
-    (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                         hi
-                         (fixnumize hi))))
-      (cond ((sc-is value descriptor-reg)
-             (let* ((mask (lognot fixnum-hi))
-                    (constant
-                      (if (= (ldb (byte 64 0) mask) non-negative-fixnum-mask-constant)
-                          (ea non-negative-fixnum-mask-constant-wired-address)
-                          (constantize mask))))
-               (inst test value constant)))
-            (t
-             (change-vop-flags vop '(:be))
-             (inst cmp value (constantize fixnum-hi)))))))
-
-(define-vop (test-fixnum-mod-tagged-unsigned)
-  (:args (value :scs (any-reg unsigned-reg signed-reg)))
-  (:arg-types (:or tagged-num unsigned-num signed-num)
-              (:constant fixnum))
-  (:translate fixnum-mod-p)
-  (:conditional :be)
-  (:info hi)
-  (:policy :fast-safe)
-  (:generator 5
-     (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                          hi
-                          (fixnumize hi))))
-       (inst cmp value (constantize fixnum-hi)))))
-
-(define-vop (test-fixnum-mod-*)
-  (:args (value :scs (any-reg descriptor-reg)))
-  (:arg-types * (:constant fixnum))
-  (:translate fixnum-mod-p)
-  (:conditional)
-  (:info target not-p hi)
-  (:policy :fast-safe)
-  (:generator 6
-     (let* ((fixnum-hi (fixnumize hi))
-            (skip (gen-label)))
-       (generate-fixnum-test value)
-       (inst jmp :ne (if not-p target skip))
-       (inst cmp value (constantize fixnum-hi))
-       (inst jmp (if not-p :a :be) target)
-       (emit-label skip))))
-
 ;;; SINGLE-FLOAT-P, CHARACTERP, UNBOUND-MARKER-P produce a flag result
 ;;; and never need a temporary.
 (macrolet ((define (name widetag)
@@ -555,7 +515,7 @@
                   out))))
   (define bignump (bignum-widetag))
   (define ratiop (ratio-widetag))
-  (define complex-rational-p (complex-widetag))
+  (define complex-rational-p (complex-rational-widetag))
   (define complex-single-float-p (complex-single-float-widetag))
   (define complex-double-float-p (complex-double-float-widetag))
   (define double-float-p (double-float-widetag))
