@@ -10,20 +10,18 @@
  */
 #include <stdio.h>
 
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 #include "runtime.h"
 #include "arch.h"
 #include "globals.h"
 #include "validate.h"
 #include "os.h"
 #include "lispregs.h"
-#include "signal.h"
-#include "alloc.h"
+#include <signal.h>
 #include "interrupt.h"
 #include "interr.h"
 #include "breakpoint.h"
-#include "monitor.h"
-#include "getallocptr.h"
+#include "pseudo-atomic.h"
 
 os_vm_address_t arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *context)
 {
@@ -32,9 +30,9 @@ os_vm_address_t arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *contex
 
 void arch_skip_instruction(os_context_t *context)
 {
-    uint32_t trap_instruction = *((uint32_t *)*os_context_pc_addr(context));
+    uint32_t trap_instruction = *(uint32_t *)OS_CONTEXT_PC(context);
     unsigned code = trap_instruction >> 5 & 0xFF;
-    *os_context_pc_addr(context) += 4;
+    OS_CONTEXT_PC(context) += 4;
     switch (code)
     {
     case trap_Error:
@@ -47,37 +45,19 @@ void arch_skip_instruction(os_context_t *context)
 
 unsigned char *arch_internal_error_arguments(os_context_t *context)
 {
-    return (unsigned char *)*os_context_pc_addr(context);
+    return (unsigned char *)OS_CONTEXT_PC(context);
 }
 
-boolean arch_pseudo_atomic_atomic(os_context_t *context)
-{
-    /* FIXME: this foreign_function_call_active test is dubious at
-     * best. If a foreign call is made in a pseudo atomic section
-     * (?) or more likely a pseudo atomic section is in a foreign
-     * call then an interrupt is executed immediately. Maybe it
-     * has to do with C code not maintaining pseudo atomic
-     * properly. MG - 2005-08-10
-     *
-     * The foreign_function_call_active used to live at each call-site
-     * to arch_pseudo_atomic_atomic, but this seems clearer.
-     * --NS 2007-05-15 */
-#ifdef LISP_FEATURE_GENCGC
-    return get_pseudo_atomic_atomic(get_sb_vm_thread());
-#else
-    return (!foreign_function_call_active)
-        && (NIL != SymbolValue(PSEUDO_ATOMIC_ATOMIC,0));
-#endif
+bool arch_pseudo_atomic_atomic(struct thread *thread) {
+    return get_pseudo_atomic_atomic(thread);
 }
 
-void arch_set_pseudo_atomic_interrupted(os_context_t *context)
-{
-    set_pseudo_atomic_interrupted(get_sb_vm_thread());
+void arch_set_pseudo_atomic_interrupted(struct thread *thread) {
+    set_pseudo_atomic_interrupted(thread);
 }
 
-void arch_clear_pseudo_atomic_interrupted(os_context_t *context)
-{
-    clear_pseudo_atomic_interrupted(get_sb_vm_thread());
+void arch_clear_pseudo_atomic_interrupted(struct thread *thread) {
+    clear_pseudo_atomic_interrupted(thread);
 }
 
 unsigned int arch_install_breakpoint(void *pc)
@@ -106,7 +86,7 @@ arch_handle_breakpoint(os_context_t *context)
 void
 arch_handle_fun_end_breakpoint(os_context_t *context)
 {
-    *os_context_pc_addr(context) = (int) handle_fun_end_breakpoint(context);
+    OS_CONTEXT_PC(context) = (uword_t) handle_fun_end_breakpoint(context);
 }
 
 void
@@ -119,20 +99,25 @@ arch_handle_single_step_trap(os_context_t *context, int trap)
 static void
 sigtrap_handler(int signal, siginfo_t *siginfo, os_context_t *context)
 {
-    uint32_t trap_instruction = *((uint32_t *)*os_context_pc_addr(context));
+    uint32_t trap_instruction = *(uint32_t *)OS_CONTEXT_PC(context);
     unsigned code = trap_instruction >> 5 & 0xFF;
     if ((trap_instruction >> 21) != 0x6A1) {
-        lose("Unrecognized trap instruction %08lx in sigtrap_handler() (PC: %p)",
-             trap_instruction, *os_context_pc_addr(context));
+        lose("Unrecognized trap instruction %08x in sigtrap_handler() (PC: %p)",
+             trap_instruction, (void*)OS_CONTEXT_PC(context));
     }
 
     handle_trap(context, code);
 }
+
+void sigill_handler(int signal, siginfo_t *siginfo, os_context_t *context);
+
+#ifndef LISP_FEATURE_DARWIN
 void
 sigill_handler(int signal, siginfo_t *siginfo, os_context_t *context) {
     fake_foreign_function_call(context);
-    lose("Unhandled SIGILL at %p.", *os_context_pc_addr(context));
+    lose("Unhandled SIGILL at %p.", (void*)OS_CONTEXT_PC(context));
 }
+#endif
 
 void arch_install_interrupt_handlers()
 {
@@ -150,8 +135,8 @@ void arch_install_interrupt_handlers()
 
 void arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
-  THREAD_JIT(0);
-  char *reloc_addr = (char*)LINKAGE_TABLE_SPACE_START + index * LINKAGE_TABLE_ENTRY_SIZE;
+  THREAD_JIT_WP(0);
+  char *reloc_addr = (char*)ALIEN_LINKAGE_TABLE_SPACE_START + index * ALIEN_LINKAGE_TABLE_ENTRY_SIZE;
 
   if (datap) {
     *(unsigned long *)reloc_addr = (unsigned long)target_addr;
@@ -181,5 +166,13 @@ void arch_write_linkage_table_entry(int index, void *target_addr, int datap)
   os_flush_icache((os_vm_address_t) reloc_addr, (char*) inst_ptr - reloc_addr);
 
  DONE:
-  THREAD_JIT(1);
+  THREAD_JIT_WP(1);
+}
+
+void gcbarrier_patch_code(void* where, int nbits)
+{
+    // Patch in the 'imms' value for UBFM
+    unsigned* pc = where;
+    unsigned int mask = ~0x0000FC00; // 6 bits at position 10
+    *pc = (*pc & mask) | ((GENCGC_CARD_SHIFT + nbits - 1) << 10);
 }
