@@ -42,7 +42,7 @@ static bool coalescible_number_p(lispobj* where)
     int widetag = widetag_of(where);
     return widetag == BIGNUM_WIDETAG
         // Ratios and complex integers containing pointers to bignums don't work.
-        || ((widetag == RATIO_WIDETAG || widetag == COMPLEX_WIDETAG)
+        || ((widetag == RATIO_WIDETAG || widetag == COMPLEX_RATIONAL_WIDETAG)
             && fixnump(where[1]) && fixnump(where[2]))
 #ifndef LISP_FEATURE_64_BIT
         || widetag == SINGLE_FLOAT_WIDETAG
@@ -215,6 +215,7 @@ void coalesce_similar_objects()
     lispobj* the_symbol_nil = (lispobj*)(NIL - LIST_POINTER_LOWTAG - N_WORD_BYTES);
     coalesce_range(the_symbol_nil, ALIGN_UP(SYMBOL_SIZE,2) + the_symbol_nil, arg);
     coalesce_range((lispobj*)STATIC_SPACE_OBJECTS_START, static_space_free_pointer, arg);
+    coalesce_range((lispobj*)PERMGEN_SPACE_START, permgen_space_free_pointer, arg);
 
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     coalesce_range((lispobj*)FIXEDOBJ_SPACE_START, fixedobj_free_pointer, arg);
