@@ -429,8 +429,6 @@ static inline void add_to_weak_pointer_chain(struct weak_pointer *wp) {
     weak_pointer_chain = wp;
 }
 
-extern unsigned char* gc_card_mark;
-
 // "assign" as the operation name is a little clearer than "set"
 // which tends to be synonymous with setting a bit to 1.
 #define assign_page_card_marks(page, val) \
@@ -479,9 +477,9 @@ extern unsigned char* gc_card_mark;
 #define NON_FAULTING_STORE(operation, addr) { \
   page_index_t page_index = find_page_index(addr); \
   if (page_index < 0 || !PAGE_WRITEPROTECTED_P(page_index)) { operation; } \
-  else { os_protect(PAGE_BASE(addr), GENCGC_PAGE_BYTES, OS_VM_PROT_JIT_ALL); \
+  else { os_protect(PAGE_BASE(addr), GENCGC_PAGE_BYTES, OS_VM_PROT_ALL); \
          operation; \
-         os_protect(PAGE_BASE(addr), GENCGC_PAGE_BYTES, OS_VM_PROT_JIT_READ); } }
+         os_protect(PAGE_BASE(addr), GENCGC_PAGE_BYTES, OS_VM_PROT_READ); } }
 
 /* This is used by the fault handler, and potentially during GC
  * if we need to remember that a pointer store occurred.
@@ -491,7 +489,7 @@ static inline void unprotect_page(void* addr, unsigned char mark)
 {
     // No atomic op needed for a 1-byte store.
     gc_card_mark[addr_to_card_index(addr)] = mark;
-    os_protect(PAGE_BASE(addr), GENCGC_PAGE_BYTES, OS_VM_PROT_JIT_ALL);
+    os_protect(PAGE_BASE(addr), GENCGC_PAGE_BYTES, OS_VM_PROT_ALL);
 }
 #endif
 
@@ -574,8 +572,6 @@ static inline os_vm_size_t npage_bytes(page_index_t npages)
 }
 extern os_vm_size_t auto_gc_trigger;
 
-void zeroize_pages_if_needed(page_index_t start, page_index_t end, int page_type);
-
 typedef unsigned int page_bytes_t;
 #define page_words_used(index) page_table[index].words_used_
 #define page_bytes_used(index) ((page_bytes_t)page_table[index].words_used_<<WORD_SHIFT)
