@@ -59,7 +59,6 @@
 #include "genesis/symbol.h"
 #include "genesis/vector.h"
 #include "immobile-space.h"
-#include "lispstring.h"
 #include "pseudo-atomic.h"
 #include "search.h"
 #include "align.h"
@@ -1633,7 +1632,7 @@ static void fixup_space(lispobj* where, size_t n_words)
         case VALUE_CELL_WIDETAG:
         case WEAK_POINTER_WIDETAG:
         case RATIO_WIDETAG:
-        case COMPLEX_WIDETAG:
+        case COMPLEX_RATIONAL_WIDETAG:
           // Use the sizing functions for generality.
           // Symbols can contain strange header bytes,
           // and vectors might have a padding word, etc.
