@@ -28,7 +28,6 @@
 #include "os.h"
 #include "core.h"
 #include "globals.h"
-#include "save.h"
 #include "dynbind.h"
 #include "lispregs.h"
 #include "validate.h"
@@ -44,6 +43,7 @@
 #ifdef LISP_FEATURE_SB_CORE_COMPRESSION
 # include <zstd.h>
 #endif
+#define COMPRESSION_LEVEL_NONE INT_MIN
 
 #define GENERAL_WRITE_FAILURE_MSG "error writing to core file"
 
@@ -307,6 +307,14 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
                  static_space_free_pointer,
                  core_start_pos,
                  core_compression_level), ++count;
+#ifdef LISP_FEATURE_PERMGEN
+    output_space(file,
+                 PERMGEN_CORE_SPACE_ID,
+                 (lispobj *)PERMGEN_SPACE_START,
+                 permgen_space_free_pointer,
+                 core_start_pos,
+                 core_compression_level), ++count;
+#endif
 #ifdef LISP_FEATURE_DARWIN_JIT
     output_space(file,
                  STATIC_CODE_CORE_SPACE_ID,
@@ -576,7 +584,6 @@ static void prepare_dynamic_space_for_final_gc(struct thread* thread)
     page_index_t i;
 
     prepare_immobile_space_for_final_gc();
-    /* TODO: external full compactor thingy for MR */
     for (i = 0; i < next_free_page; i++) {
 #ifndef LISP_FEATURE_MARK_REGION_GC
         // Compaction requires that we permit large objects to be copied henceforth.
@@ -718,6 +725,16 @@ gc_and_save(char *filename, bool prepend_runtime, bool purify,
     struct thread *thread = get_sb_vm_thread();
     gc_close_thread_regions(thread, 0);
     gc_close_collector_regions(0);
+    unwind_binding_stack(thread);
+    // Avoid tenuring of otherwise-dead objects referenced by bindings which
+    // disappear on image restart. This must occcur *after* unwind_binding_stack()
+    // because unwinding moves values from the binding stack into TLS.
+    char *start = (char*)&thread->lisp_thread;
+    char *end = (char*)thread + dynamic_values_bytes;
+    memset(start, 0, end-start);
+    // After zeroing, make sure PINNED_OBJECTS is a list again.
+    write_TLS(PINNED_OBJECTS, NIL, thread);
+
     pre_verify_gen_0 = 1;
 #ifdef LISP_FEATURE_MARK_REGION_GC
     /* Do a minor GC to instate allocation bitmap for new objects.
@@ -728,15 +745,6 @@ gc_and_save(char *filename, bool prepend_runtime, bool purify,
     move_rospace_to_dynamic(0);
     prepare_immobile_space_for_final_gc(); // once is enough
     prepare_dynamic_space_for_final_gc(thread);
-    unwind_binding_stack(thread);
-    // Avoid tenuring of otherwise-dead objects referenced by bindings which
-    // disappear on image restart. This must occcur *after* unwind_binding_stack()
-    // because unwinding moves values from the binding stack into TLS.
-    char *start = (char*)&thread->lisp_thread;
-    char *end = (char*)thread + dynamic_values_bytes;
-    memset(start, 0, end-start);
-    // After zeroing, make sure PINNED_OBJECTS is a list again.
-    write_TLS(PINNED_OBJECTS, NIL, thread);
 
     save_lisp_gc_iteration = 1;
 #ifndef LISP_FEATURE_MARK_REGION_GC
